/*
 * This file is part of CityLifeAI, an AI for OpenTTD.
 *
 * It's free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the
 * Free Software Foundation, version 2 of the License.
 *
 */

class Town
{
    id = null;                      // Town id
    depot = null;                   // Built depo
    vehicle_group = null;           // Group ID of this town vehicles
    vehicle_list = null;            // List of owned vehicles
    population = null;              // Monthly population count 
    pax_transported = null;         // Monthly percentage of transported pax
    mail_transported = null;        // Monthly percentage of transported mail
    connections = null;             // List of established road connections
	initialized = null;
	
	constructor(town_id, load_town_data=false)
	{	
		this.vehicle_list = [];
		this.population = 0;
		this.pax_transported = 0;
		this.mail_transported = 0;
		this.connections = [];	
		this.id = town_id;
		this.initialized = false;
		UpdateStats()
		
		/* If there isn't saved data for the towns, we
		 * initialize them. Otherwise, we load saved data.
		 */
		if (!load_town_data)
		{
			this.vehicle_group = AIGroup.CreateGroup(AIVehicle.VT_ROAD, AIGroup.GROUP_INVALID);
			AIGroup.SetName(this.vehicle_group, AITown.GetName(this.id));
		}
		else
		{
			this.depot = ::TownDataTable[this.id].depot;
			this.initialized = (this.depot != null)
			this.vehicle_group = ::TownDataTable[this.id].vehicle_group;
			this.connections = ::TownDataTable[this.id].connections;

			// Recreate list of vehicles from group information
			// What was the point of saving and loading sell status when it will be safer re-applied on next town check?
			// In fact, re-applying fixes the bug of stuck cars, while save-loading keeps it
			if (AIGroup.IsValidGroup(this.vehicle_group))
			{
				local vehicles_in_group = AIVehicleList_Group(this.vehicle_group);
				foreach (vehicle, _ in vehicles_in_group)
					this.vehicle_list.append(Vehicle(vehicle, ::EngineList.GetValue(AIVehicle.GetEngineType(vehicle))))
				AILog.Info("Total vecs in town: " + AITown.GetName(this.id) + " " + this.vehicle_list.len());
			}
		}
	}
}

function Town::SaveTownData()
{
    local town_data = {};
    town_data.depot <- this.depot;
    town_data.vehicle_group <- this.vehicle_group;
    town_data.connections <- this.connections;

    return town_data;
}

function Town::InitTown()
{
    if (::EngineList.Count() == 0)
        return;

    if (this.depot == null)
    {
        AILog.Info("Trying to build a depot in town " + AITown.GetName(this.id));
        this.depot = BuildDepot(this.id);
		this.initialized = (this.depot != null)
    }	
}

//I know that's not great, via shr/shl better, don't care
function Sqrt(x){
	//x		1	16	64	144		256		400		625		10000	1000000		100000000
	//ret	1	4	8	12		16		20		25		100		1000		10000
	//
	if		(x < 1)			return	x
	else if (x < 16)		return  1 		+ 	3	*(x-1)			/(16-1)
	else if (x < 64)		return  4 		+ 	4	*(x-16)			/(64-16)
	else if (x < 144)		return  8 		+ 	4	*(x-64)			/(144-64)
	else if (x < 256)		return  12 		+ 	4	*(x-144)		/(256-144)
	else if (x < 400)		return  16 		+ 	4	*(x-256)		/(400-256)
	else if (x < 625)		return  20 		+ 	5	*(x-400)		/(625-400)		
	else if (x < 10000)		return  25 		+ 	75	*(x-625)		/(10000-625)		
	else if (x < 1000000)	return  100 	+ 	900	*(x-10000)		/(1000000-10000)
	else if (x < 100000000)	return  1000 	+ 	9000*(x-1000000)	/(100000000-1000000)
	else					return  10000 	+ 	x/10000
}

function Town::DailyManageTown()
{
    if ((::EngineList.Count() == 0) || !this.initialized)
        return;

	local sane_population  = Sqrt(this.population)
	
	local personal_count = ceil(sane_population / 10.0 * CalculateVehicleCountDecrease(this.pax_transported, 30))

	local service_count = floor(sane_population / 20.0 * CalculateVehicleCountDecrease(this.mail_transported, 30, 80))
	if (GetEngineListByCategory(Category.MAIL | Category.GARBAGE).Count() > 0)
		ManageVehiclesByCategory(service_count, Category.MAIL | Category.GARBAGE)
	else
		personal_count += service_count

	local emergency_count = floor(sane_population / 25.0)
	if (GetEngineListByCategory(Category.FIRE | Category.POLICE | Category.AMBULANCE).Count() > 0)
		ManageVehiclesByCategory(emergency_count, Category.FIRE | Category.POLICE | Category.AMBULANCE);
	else
		personal_count += emergency_count

	ManageVehiclesByCategory(personal_count, Category.CAR);
	UpdateVehicles();    
}

function Town::UpdateStats() {
    this.population = AITown.GetPopulation(this.id)
    this.pax_transported = AITown.GetLastMonthTransportedPercentage(this.id, 0x00);
	this.mail_transported = AITown.GetLastMonthTransportedPercentage(this.id, 0x02);
}

function Town::SellObsoletes() {
	for (local i = 0; i < this.vehicle_list.len(); i++)
	{
		//Avoiding duplications using yearly updated ::EngineList
		vehicle_list[i].category = ::EngineList.GetValue(AIVehicle.GetEngineType(vehicle_list[i].id))
		if (vehicle_list[i].category & Category.OBSOLETE)
			vehicle_list[i].Sell();
	}
}

function Town::MonthlyManageTown()
{
	if (!this.initialized)
		return
	UpdateStats()	
	SellObsoletes()

    // // TODO: Remove
    // local personal_count = ceil(this.population / 100.0 * this.CalculateVehicleCountDecrease(this.pax_transported, 30));
    // local service_count = ceil(this.population / 500.0 * this.CalculateVehicleCountDecrease(this.mail_transported, 30, 80));
    // local emergency_count = ceil((this.population - 1000) / 2000.0) * 3;
    // AILog.Info(AITown.GetName(this.id) + ": Population = " + this.population + ", Pax transported = " + this.pax_transported + " Mail transported = " + this.mail_transported);
    // AILog.Info("Personal = " + personal_count + ", Services = " + service_count + ", Emergency = " + emergency_count);
}

/*function Town::DebugVecs(){
	AILog.Info("Total vecs in town: " + AITown.GetName(this.id) + " " + this.vehicle_list.len());
	foreach(vec in this.vehicle_list)
		AILog.Info(vec);
}*/


function Town::ManageVehiclesByCategory(target_count, category)
{
    local vehicle_count = GetVehicleCountByCategory(this.vehicle_list, category);
	//DebugVecs()
	
    AILog.Info(AITown.GetName(this.id) + ": " + category + " (" + vehicle_count + "/" + target_count + ")");
    if (vehicle_count > target_count + 1)
    {
        AILog.Info("Selling " + (vehicle_count - target_count) + " vehicles of type " + category);
        local to_sell_list = ExtractByCategory(this.vehicle_list, category);
        for (local i = 0; i < vehicle_count - target_count; ++i)        
            to_sell_list[i].Sell()

    }
    else if (vehicle_count < target_count)
    {
        local company_vehicles_count = AIVehicleList().Count();
        local max_vehicles = AIGameSettings.GetValue("max_roadveh");

        AILog.Info("Buying " + (target_count - vehicle_count) + " vehicles of type " + category);
        local engine_list = GetEngineListByCategory(category, Category.OBSOLETE)
		if (engine_list.Count() == 0)
			return;

        // Randomize start of the vehicle list
        local engine = engine_list.Begin();
        local rand = AIBase.RandRange(engine_list.Count());
        for (local i = 0; i < rand; ++i)
            engine = engine_list.Next();


        for (local i = 0; (i < target_count - vehicle_count) && (company_vehicles_count + i < max_vehicles); ++i)
        {
            local vehicle = AIVehicle.BuildVehicle(this.depot, engine);
            if (AIVehicle.IsValidVehicle(vehicle))
            {
                this.vehicle_list.append(Vehicle(vehicle, engine_list.GetValue(engine)));
                AIGroup.MoveVehicle(this.vehicle_group, vehicle);
            }
            else
            {
                break;
            }

            engine = engine_list.Next();
            if (engine_list.IsEnd())
                engine = engine_list.Begin();
        }
    }
}

function Town::CalculateVehicleCountDecrease(transported, min_transported, max_transported=100)
{
    if (transported < min_transported)
        return 1.0
    else if (transported > max_transported)
        return 0.0
    else
        return (1.0 - (transported - min_transported).tofloat() / (max_transported - min_transported).tofloat())
}

function Town::UpdateVehicles(){
    for (local i = 0; i < this.vehicle_list.len(); ++i)
        if (this.vehicle_list[i].Update())
            this.vehicle_list.remove(i--)
}

function Town::RemoveVehicle(vehicle_id){
    for (local i = 0; i < this.vehicle_list.len(); ++i)
        if (this.vehicle_list[i].id == vehicle_id){
            this.vehicle_list.remove(i)
            return true
        }

    return false;
}

function Town::Parade(town_b)
{
    local company_vehicles_count = AIVehicleList().Count();
    local max_vehicles = AIGameSettings.GetValue("max_roadveh");

    local engine_list = GetEngineListByCategory(Category.SPORT);
    if (engine_list.Count() == 0)
        engine_list = AIEngineList(AIVehicle.VT_ROAD);
        engine_list.Valuate(AIEngine.GetMaxSpeed);
        engine_list.KeepTop(1);

    local engine = engine_list.Begin();
    for (local i = 0; (i < 10) && (company_vehicles_count + i < max_vehicles); ++i)
    {
        local purchased = AIVehicle.BuildVehicle(this.depot, engine);
        if (AIVehicle.IsValidVehicle(purchased))
        {
            local vehicle = Vehicle(purchased, engine_list.GetValue(engine));
            vehicle.action = Action.SELL;
            this.vehicle_list.append(vehicle);
            AIVehicle.StartStopVehicle(purchased);
            AIGroup.MoveVehicle(this.vehicle_group, purchased);
            AIOrder.AppendOrder(purchased, town_b.depot, AIOrder.OF_STOP_IN_DEPOT);
        }
        else
        {
            break;
        }

        engine = engine_list.Next();
        if (engine_list.IsEnd())
            engine = engine_list.Begin();
    }
}